import { getContextItem } from "../lib/context.ts";
import { Doodler, Vector } from "@bearmetal/doodler";
import { Spline, TrackSegment, TrackSystem } from "../track/system.ts";
import { Debuggable } from "../lib/debuggable.ts";
import { map } from "../math/lerp.ts";
import { off } from "node:process";

export class Train extends Debuggable {
  nodes: Vector[] = [];

  cars: TrainCar[] = [];

  path: Spline<TrackSegment>;
  t: number;

  spacing = 20;

  speed = 0;

  get segments() {
    return Array.from(
      new Set(this.cars.flatMap((c) => c.segments.values().toArray())),
    );
  }

  constructor(track: Spline<TrackSegment>, cars: TrainCar[]) {
    super("train", "path");
    this.path = track;
    this.t = 0;
    this.cars = cars;

    let currentOffset = 0;
    try {
      for (const car of this.cars) {
        car.train = this;
        currentOffset += car.moveAlongPath(this.t - currentOffset) +
          this.spacing;
      }
      console.log("forward");
    } catch {
      currentOffset = 0;
      console.log("Reversed");
      for (const car of this.cars.toReversed()) {
        for (const [i, bogie] of car.bogies.entries().toArray().reverse()) {
          currentOffset += bogie.length;
          const a = this.path.followEvenPoints(this.t - currentOffset);
          car.setBogiePosition(a.p, i);
          this.nodes.push(a.p);
          car.segments.add(a.segmentId);
        }
      }
    }
  }

  move(dTime: number) {
    if (!this.speed) return;
    this.t = this.t + this.speed * dTime * 10;
    // % this.path.evenPoints.length; // This should probably be on the track system
    let currentOffset = 0;
    for (const car of this.cars) {
      // This needs to be moved to the car itself
      // if (!car.points) return;
      // const [a, b] = car.points;
      // const nA = this.path.followEvenPoints(this.t - currentOffset);
      // a.set(nA.p);
      // currentOffset += car.length;
      // const nB = this.path.followEvenPoints(this.t - currentOffset);
      // b.set(nB.p);
      // currentOffset += this.spacing;
      // car.segments = [nA.segmentId, nB.segmentId];
      // car.draw();

      currentOffset += car.moveAlongPath(this.t - currentOffset) + this.spacing;
    }
    // this.draw();
  }

  // draw() {
  //   const doodler = getContextItem<Doodler>("doodler");
  //   this.path.draw();
  //   for (const [i, node] of this.nodes.entries()) {
  //     // doodler.drawCircle(node, 10, { color: "purple", weight: 3 });
  //     doodler.fillCircle(node, 2, { color: "purple" });
  //     // const next = this.nodes[i + 1];
  //     // if (next) {
  //     //   const to = Vector.sub(node.point, next.point);
  //     //   to.setMag(40);
  //     //   doodler.line(next.point, Vector.add(to, next.point))
  //     // }
  //   }
  // }

  draw() {
    for (const car of this.cars) {
      car.draw();
    }
  }

  override debugDraw(): void {
    const debug = getContextItem<Debug>("debug");
    const doodler = getContextItem<Doodler>("doodler");
    if (debug.path) {
      // doodler.drawLine(this.path.points, {
      //   color: "red",
      //   weight: 3,
      // });
      const colors = getContextItem<string[]>("colors");
      for (const [i, p] of this.path.evenPoints.entries()) {
        const color = colors[
          Math.floor(
            map(i, 0, this.path.evenPoints.length, 0, colors.length),
          )
        ];
        doodler.drawCircle(p.p, 2, { color, weight: .5 });
      }
    }

    if (debug.train) {
      const track = getContextItem<TrackSystem>("track");
      const colors = getContextItem<string[]>("colors").slice();
      colors.push(colors.shift()!);
      colors.push(colors.shift()!);
      colors.push(colors.shift()!);
      for (const [i, segmentId] of this.segments.entries().toArray()) {
        const segment = track.getSegment(segmentId);
        segment &&
          doodler.drawBezier(...segment.points, {
            color: colors[i % colors.length],
            weight: 3,
          });
      }
    }
  }

  real2Track(length: number) {
    return length / this.path.pointSpacing;
  }
}

interface Bogie {
  pos: Vector;
  angle: number;
  length: number;
}

export class TrainCar extends Debuggable {
  img: HTMLImageElement;
  imgWidth: number;
  imgHeight: number;
  sprite?: ISprite;

  points?: [Vector, Vector, ...Vector[]];
  length: number;

  bogies: Bogie[] = [];

  segments: Set<string> = new Set();

  train?: Train;

  constructor(
    length: number,
    trailing: number,
    img: HTMLImageElement,
    w: number,
    h: number,
    sprite?: ISprite,
  ) {
    super(true, "car");
    this.img = img;
    this.sprite = sprite;
    this.imgWidth = w;
    this.imgHeight = h;
    this.length = length;

    this.bogies = [
      {
        pos: new Vector(0, 0),
        angle: 0,
        length: length,
      },
      {
        pos: new Vector(0, 0),
        angle: 0,
        length: trailing,
      },
    ];
  }

  setBogiePosition(pos: Vector, idx: number) {
    this.bogies[idx].pos.set(pos);
  }

  update(dTime: number, t: number) {
    if (this.train) {
      for (const [i, bogie] of this.bogies.entries()) {
        const a = this.train.path.followEvenPoints(t - this.length * i);
      }
    }
  }

  moveAlongPath(t: number): number {
    if (!this.train) return 0;
    let offset = 0;
    this.segments.clear();
    for (const [i, bogie] of this.bogies.entries()) {
      const a = this.train.path.followEvenPoints(t - offset);
      offset += bogie.length;
      this.setBogiePosition(a.p, i);
      this.segments.add(a.segmentId);
    }
    return offset;
  }

  draw() {
    const doodler = getContextItem<Doodler>("doodler");
    const [a, b] = this.bogies;
    const origin = Vector.add(Vector.sub(a.pos, b.pos).div(2), b.pos);
    const angle = Vector.sub(b.pos, a.pos).heading();

    doodler.drawCircle(origin, 4, { color: "blue" });

    doodler.drawRotated(origin, angle, () => {
      this.sprite
        ? doodler.drawSprite(
          this.img,
          this.sprite.at,
          this.sprite.width,
          this.sprite.height,
          origin.copy().sub(this.imgWidth / 2, this.imgHeight / 2),
          this.imgWidth,
          this.imgHeight,
        )
        : doodler.drawImage(
          this.img,
          origin.copy().sub(this.imgWidth / 2, this.imgHeight / 2),
        );
    });
  }
  override debugDraw(...args: unknown[]): void {
    const doodler = getContextItem<Doodler>("doodler");
    doodler.drawLine(this.bogies.map((b) => b.pos), {
      color: "blue",
      weight: 2,
    });
    doodler.deferDrawing(() => {
      const colors = getContextItem<string[]>("colors");
      for (const [i, b] of this.bogies.entries()) {
        doodler.drawCircle(b.pos, 5, { color: colors[i % colors.length] });
        doodler.fillText(b.length.toString(), b.pos.copy().add(10, 0), 100, {
          color: "white",
        });
      }
    });
  }
}

interface ISprite {
  at: Vector;
  width: number;
  height: number;
}
